#!/usr/bin/env python

import sys
import os

import sqlite3
import codecs
import time
import hashlib


version = "20131102-3"

# Utility functions and stuff
################################################################################

def listdir(path):
    # Make os.listdir return unicode names by passing it a unicode path
    if isinstance(path, bytes):
        path = path.decode(sys.getfilesystemencoding())
    return os.listdir(path)

_case_sensitive = ('abcDEF' == os.path.normcase('abcDEF'))
def realname(path):
    """ Return true if the name is the real name, else return false. """

    if _case_sensitive:
        return True

    # On case insensitive systems, we want to make sure the name matches
    # what the OS tells us it would be from a directory listing.
    dir = os.path.dirname(path)
    name = os.path.basename(path)

    return name in listdir(dir)

def checksum(path):
    """ Calculate the checksum and return the result. """
    hasher = hashlib.md5()

    with open(path, 'rb') as handle:
        data = handle.read(4096000)
        while len(data):
            hasher.update(data)
            data = handle.read(4096000)

    return hasher.hexdigest()

def output(msg):
    sys.stdout.write(msg + '\n')
    sys.stdout.flush()


def message(msg):
    sys.stderr.write(msg + '\n');
    sys.stderr.flush()
    
def status(path, status):
    message(status + ': ' + path)

_verbose = False
def verbose(path, status):
    if _verbose:
        status(msg, status)


# Collection classes
################################################################################
class Collection(object):

    TYPE_DIRECTORY = 0
    TYPE_FILE = 1
    TYPE_SYMLINK = 2

    def __init__(self, root):
        self.root = root
        self.db = sqlite3.connect(root + os.sep + 'collection.db')
        self.db.execute('PRAGMA foreign_keys = ON;')

    def create(self):
        self.db.execute('''CREATE TABLE tree(
            id INTEGER PRIMARY KEY NOT NULL,
            parent INTEGER,
            name TEXT,
            type INTEGER,
            FOREIGN KEY (parent) REFERENCES tree (id) ON DELETE CASCADE,
            UNIQUE (name, parent)
        )''')

        self.db.execute('''CREATE INDEX idx_tree_name ON tree (name)''')

        self.db.execute('''CREATE TABLE symlink(
            id INTEGER PRIMARY KEY NOT NULL,
            treeitem INTEGER,
            target TEXT,
            FOREIGN KEY (treeitem) REFERENCES tree (id) ON DELETE CASCADE
        )''')

        self.db.execute('''CREATE TABLE file(
            id INTEGER PRIMARY KEY NOT NULL,
            treeitem INTEGER,
            checksum TEXT,
            size INTEGER,
            timestamp INTEGER,
            FOREIGN KEY (treeitem) REFERENCES tree (id) ON DELETE CASCADE
        )''')

    def _check_helper(self, full, path, id)
        names = []
        result = True

        # Check for missing items
        cur = self.db.cursor()
        cur.execute('SELECT id,name,type FROM tree WHERE parent = ?', (id,))

        row = cur.fetchone()
        while row != None:
            if row[2] == self.TYPE_DIRECTORY:
                result = result and self._check_directory(full, path + os.sep + row[1], row[0])
            elif row[2] == self.TYPE_FILE:
                result = result and self._check_file(full, path + os.sep + row[1], row[0])
            elif row[2] == self.TYPE_SYMLINK:
                result = result and  and self._check_symlink(full, path + os.sep + row[1], row[0])
                
            names.append(row[1])
            row = cur.fetchone()

        # Check for new items
        realpath = self.root + os.sep + path
        if isinstance(realpath, bytes):
            realpath = realpath.decode(sys.getfilesystemencoding())

        for i = os.listdir(realpath):
            if not i in names:
                result = False
                # New item
                # If directory, show al child items as well


    def _check_directory(self, full, path, id):
        realpath = self.root + os.sep + path
        result = True

        if not (os.path.isdir(realpath) and not os.path.islink(realpath)):
            status(path, 'MISSING')
            result = False

        return result and self._check_helper(self, full, path, id)

    def _check_file(self, full, path, id):
        pass

    def _check_symlink(self, full, path, id):
        pass


    def check(self, full):
        return self._check_helper(full, '', 0)

    def update(self):
        pass

    def dump(self):
        pass


# Program entry point
################################################################################
def usage():
    print('Usage: ' + sys.argv[0] + ' [-v] <action>')
    print('')
    print('  -v:        Verbose status messages')
    print('  action:    create, check, verify, update, dump')
    print('')

def main():
    # Check arguments
    if len(sys.argv) < 2:
        usage()
        sys.exit(-1)

    if sys.argv[1] == '-v':
        if len(sys.argv) != 3:
            usage()
            sys.exit(-1)

        verbose = True
        action = sys.argv[2]
    else:
        verbose = False
        action = sys.argv[1]

    if not action in ('create', 'check', 'verify', 'update', 'dump'):
        usage()
        sys.exit(-1)

    # Do stuff
    root = os.getcwd()

    if action == 'create':
        coll = Collection(root)
        coll.create()
    elif action == 'check':
        coll = Collection.load(root)
        if not coll.check(log):
            sys.exit(-1)
    elif action == 'verify':
        coll = Collection.load(root)
        if not coll.check(log, True):
            sys.exit(-1)
    elif action == 'update':
        coll = Collection.load(root)
        coll.update(log)
        coll.save()
    elif action == 'dump':
        coll = Collection.load(root)
        coll.dumpchecksum(log)

if __name__ == '__main__':
    main()

